/* User privilege/validation checking.
   Copyright (c) 1999, 2000 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>

#include "libvsd.h"

/* Get minimum and maximum uid and gid ranges for a virtual server.
   If server_root is NULL then assume we are chroot to the VS.
   Accept uid/gid pointers as NULL, returning values only for those
   pointers that are not NULL (obviously).
   Return 0 on success, 1 on failure.  */
int vsd_get_uidgid (const char *server_root,
		    uid_t *uid_min, uid_t *uid_max,
		    gid_t *gid_min, gid_t *gid_max)
{
  struct vsd_file_line *logindefs, *line;
  char *p, key[128];
  int x;

  if (! server_root)
    server_root = "/";

  if (vsd_load_file (&logindefs, server_root, "/etc/login.defs"))
    return 1;

  for (line = logindefs; line; line = line->next)
    {
      /* Skip white space.  */
      for (p = line->line; isspace (*p); p++)
	;

      /* Ignore comment lines and skip empty lines.  */
      if (*p == '#' || *p == '\0')
	continue;

      for (x = 0; ! isspace (*p); x++, p++)
	key[x] = *p;
      key[x] = '\0';

      if (uid_min && ! strcmp (key, "UID_MIN"))
	*uid_min = (uid_t) atoi (p);
      if (uid_max && ! strcmp (key, "UID_MAX"))
	*uid_max = (uid_t) atoi (p);
      if (gid_min && ! strcmp (key, "GID_MIN"))
	*gid_min = (gid_t) atoi (p);
      if (gid_max && ! strcmp (key, "GID_MAX"))
	*gid_max = (gid_t) atoi (p);
    }
  vsd_free_file (logindefs);
  return 0;
}

/* Is `uid' within the range available to the virtual server.
   Return -1 on error, 0 on failure and 1 on success.  */
int vsd_valid_uid (uid_t uid)
{
  uid_t uid_min, uid_max;

  if (vsd_get_uidgid (NULL, &uid_min, &uid_max, NULL, NULL))
    return -1;

  if (uid < uid_min || uid > uid_max)
    {
      errno = ERANGE;
      return 0;
    }

  return 1;
}

/* Is `gid' within the range available to the virtual server.
   Return -1 on error, 0 on failure and 1 on success.  */
int vsd_valid_gid (gid_t gid)
{
  gid_t gid_min, gid_max;

  if (vsd_get_uidgid (NULL, NULL, NULL, &gid_min, &gid_max))
    return -1;

  if (gid < gid_min || gid > gid_max)
    {
      errno = ERANGE;
      return 0;
    }

  return 1;
}

/* Check uid is a valid user on the virtual server.
   Return 1 if valid, 0 if not valid and -1 on error.  */
int vsd_validate_user_uid (uid_t uid)
{
  struct passwd *pw;

  if (uid == 0) /* Allow root user */
    return 1;

  /* Check user is defined in the password list.  */
  pw = getpwuid (uid);
  if (pw == NULL)
    {
      errno = ENOENT;
      return 0;
    }

  /* Check uid is in the range defined for the virtual server.  */
  return vsd_valid_uid (uid);
}

/* Check `username' is a valid user on the virtual server.
   Return 1 if valid, 0 if not valid and -1 on error.  */
int vsd_validate_user_name (const char *username)
{
  struct passwd *pw;

  if (strcmp (username, "root") == 0)
    return 1;

  if (username == NULL)
    {
      errno = EINVAL;
      return -1;
    }

  pw = getpwnam (username);
  if (pw == NULL)
    {
      errno = ENOENT;

      return 0;
    }

  return vsd_valid_uid (pw->pw_uid);
}

/* Return 1 if `uid' belongs to the admin user. 0 if not.  */
int vsd_admin_user (uid_t uid)
{
  struct passwd *pw;

  pw = getpwuid (uid);
  if (pw == NULL)
    {
      errno = ENOENT;
      return -1;
    }

  if (strcmp (pw->pw_name, "admin") == 0)
    return 1;

  return 0;
}

/* Return 0 on success, 1 on failure, -1 on error.  */
int vsd_check_path (const char *path, int check_file, int access_rights)
{
  struct stat b;
  char *tmp, *p;

  if (check_file)
    if (access (path, access_rights))
      {
	if (errno == EACCES)
	  {
	    if (stat (path, &b))
	      return 1;
	    
	    if (vsd_valid_uid (b.st_uid) == 1
		&& vsd_valid_gid (b.st_gid) == 1)
	      return 0;
	  }
	
	if (errno != ENOENT)
	  /* Access denied.  */
	  return 1;
      }

  /* Destination exists and we are allowed to write to it.  */

  /* Strip file from path.  */
  tmp = alloca (strlen (path) + 1);

  strcpy (tmp, path);
  p = strrchr (tmp, '/');
  if (p == NULL)
    strcpy (tmp, "./");
  else
    *++p = '\0';

  if (access (tmp, access_rights))
    {
      if (errno == EACCES)
	{
	  if (stat (tmp, &b))
	    return 1;
	  /* access says that we don't (as the ordinary user) have permission
	     to access the path. But we say the admin user should be granted
	     permission provided the destination uid/gid is within the
	     ranges of his virtual server.  */
	  if (vsd_valid_uid (b.st_uid) == 1 && vsd_valid_gid (b.st_gid) == 1)
	    return 0;
	}

      /* Access denied.  */
      return 1;
    }

  /* Access granted.  */
  return 0;
}

/* Is pid a member of our Virtual Server ?
   Return 1 if yes, 0 if no, -1 on error.  */
int vsd_owner_pid (pid_t pid)
{
  char tmp[128], buff[256];

  /* If the process is not a member of our virtual server then
     readlink will return a permission denied error.  */
  sprintf (tmp, "/proc/%d/root", pid);
  memset (buff, 0, sizeof (buff));
  if (readlink (tmp, buff, sizeof (buff)) == -1)
    return -1;

  if (buff[0] == '/' && buff[1] == '\0')
    return 1;

  return 0;
}

/* Return NULL if the username valid.  Otherwise return a string
   indicating the error.  */
const char *vsd_valid_username (const char *user)
{
  if (user == NULL)
    return "is_username_valid: passed NULL argument";

  if (strspn (user, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-") != strlen (user))
    return "contains unsuitable characters";

  if (isdigit (user[0]))
    return "cannot start with a digit";

  /* The shadow password suite restricts us to 16 characters.  */
  if (strlen (user) > 16)
    return "maximum length is 16 characters";

  return NULL;
}

/* Return NULL if the username is valid.  Otherwise return a string
   indicating the error.  */
const char *vsd_valid_vsname (const char *vs)
{
  if (vs == NULL)
    return "is_vsname_valid: passed NULL argument";

  if (strspn (vs, "abcdefghijklmnopqrstuvwxyz0123456789") != strlen (vs))
    return "contains unsuitable characters";

  if (isdigit (vs[0]))
    return "cannot start with a digit";

  /* Restrict virtual server names to 20 characters.  */
  if (strlen (vs) > 20)
    return "name is too long";

  return NULL;
}
