/* Apache configuration file manipulation.
   Copyright (c) 1999, 2000 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>

#include "vsd.h"
#include "libvsd.h"

/* Should be moved to config.h and configure.in.  */
#define HTTPD_CONF "/etc/httpd/conf/httpd.conf"

static char **scan_virtual_hosts (struct vsd_file_line *head, int *total);
static int delete_virtual_host (struct vsd_file_line *head,
				const char *virtual_host,
				const char *server_name);
static int add_virtual_host (struct connection *vc,
			     struct vsd_file_line *head, const char *data);
static struct vsd_file_line *find_virtual_host (struct vsd_file_line *head,
						const char *server,
						const char *servername);
static int check_vh_config (struct connection *vc, struct vsd_file_line *line,
			    struct vsd_file_line *last);

/* Simplify a path name in place, deleting redundant components.  This
   reduces OS overhead and guarantees that equivalent paths compare
   the same (modulo symlinks).

   Transforms made:
   foo/bar/../quux      foo/quux
   foo/./bar            foo/bar
   foo//bar             foo/bar
   /../quux             /quux
   //quux               //quux  (POSIX allows leading // as a namespace escape)

   Guarantees no trailing slashes. All transforms reduce the length
   of the string.

   This function is shamelessly taken from cpplib in the GNU GCC 2.96
   development source tree.  */
static void simplify_pathname (char *path)
{
  char *from = path, *to = path;
  char *base;
  int absolute = 0;

  /* Remove redundant initial /s.  */
  if (*from == '/')
    {
      absolute = 1;
      to++;
      from++;
      if (*from == '/')
        {
	  if (*++from == '/')
	    /* 3 or more initial /s are equivalent to 1 /.  */
	    while (*++from == '/');
	  else
	    /* On some hosts // differs from /; Posix allows this.  */
	    to++;
        }
    }
  base = to;
  
  for (;;)
    {
      while (*from == '/')
	from++;
      
      if (from[0] == '.' && from[1] == '/')
	from += 2;
      else if (from[0] == '.' && from[1] == '\0')
	goto done;
      else if (from[0] == '.' && from[1] == '.' && from[2] == '/')
        {
	  if (base == to)
            {
	      if (absolute)
		from += 3;
	      else
                {
		  *to++ = *from++;
		  *to++ = *from++;
		  *to++ = *from++;
		  base = to;
                }
            }
	  else
            {
	      to -= 2;
	      while (to > base && *to != '/') to--;
	      if (*to == '/')
		to++;
	      from += 3;
            }
        }
      else if (from[0] == '.' && from[1] == '.' && from[2] == '\0')
        {
	  if (base == to)
            {
	      if (!absolute)
                {
		  *to++ = *from++;
		  *to++ = *from++;
                }
            }
	  else
            {
	      to -= 2;
	      while (to > base && *to != '/') to--;
	      if (*to == '/')
		to++;
            }
	  goto done;
        }
      else
	/* Copy this component and trailing /, if any.  */
	while ((*to++ = *from++) != '/')
	  if (!to[-1])
	    {
	      to--;
	      goto done;
	    }
    }

 done:
  /* Trim trailing slash */
  if (to[0] == '/' && (!absolute || to > path+1))
    to--;
  
  /* Change the empty string to "." so that stat() on the result
     will always work. */
  if (to == path)
    *to++ = '.';
  
  *to = '\0';
}

/* Look in `line' for a token that starts with the characters in `directive'
   and ends with `terminator'.  If found, return a malloc`d string consisting
   of the characters between the `directive' and the `terminator'.  Return
   NULL if not found.  */
static char *scan_directive (const char *line, const char *directive,
			     char terminator)
{
  int count;
  const char *p;
  char *buffer;

  line = vsd_file_skip_spaces (line);

  if (strncmp (line, directive, strlen (directive)) != 0)
    return NULL;

  /* Skip past directive.  */
  line += strlen (directive);
  line = vsd_file_skip_spaces (line);

  /* Determine length of directive argument.  */
  for (count = 0, p = line; *p && *p != terminator; count++, p++)
    ;

  /* Strip trailing spaces.  */
  while (*p == terminator || *p == ' ')
    p--, count--;

  buffer = (char *) malloc (count + 2);
  strncpy (buffer, line, count + 1);
  buffer[count + 1] = '\0';

  return buffer;
} 

/* Fill `host_list' with a vector of virtual host entries.  */
static char **scan_virtual_hosts (struct vsd_file_line *head, int *total)
{
  struct vsd_file_line *entry;
  int x;
  char *vhost, *sname, **host_list;

  *total = 0;
  entry = head;
  while (entry)
    {
      if (strncmp (vsd_file_skip_spaces (entry->line),
		   "<VirtualHost", 12) == 0)
	*total = *total + 1;
      entry = entry->next;
    }

  if (*total == 0)
    return NULL;

  host_list = (char **) malloc (((*total) + 1) * sizeof (char *));
  x = 0;
  entry = head;
  while (entry)
    {
      if ((vhost = scan_directive (entry->line, "<VirtualHost", '>')))
	{
	  int found = 0;

	  while (entry)
	    {
	      if ((sname = scan_directive (entry->line, "ServerName", '\n')))
		{
		  host_list[x] = (char *) malloc (strlen (vhost)
						  + strlen (sname) + 12);
		  sprintf (host_list[x], "vh=\"%s\" sn=\"%s\"", vhost, sname);
		  x++;
		  found = 1;
		  free (sname);
		  break;
		}
	      if (! strncmp (vsd_file_skip_spaces (entry->line),
			     "</VirtualHost>", 14))
		break;

	      entry = entry->next;
	    }
	  if (! found)
	    {
	      host_list[x] = (char *) malloc (12 + strlen (vhost) + 1);
	      sprintf (host_list[x], "%s;unknown", vhost);
	      x++;
	    }

	  free (vhost);
	}
      entry = entry->next;
    }

  return host_list;
}

static int delete_virtual_host (struct vsd_file_line *head,
				const char *virtual_host,
				const char *servername) 
{
  struct vsd_file_line *host_start, *host_end, *link;

  /* Point `host_start' to the start of the VirtualHost declaration.  */
  host_start = find_virtual_host (head, virtual_host, servername);
  if (host_start == NULL)
    return -1;

  /* Point `host_end' to the end of the VirtualHost declaration.  */
  host_end = host_start;
  while (host_end)
    {
      if (strncmp (vsd_file_skip_spaces (host_end->line),
		  "</VirtualHost>", 14) == 0)
	break;

      host_end = host_end->next;
    }

  if (host_end == NULL)
    {
      /* Closing </VirtualHost> couldn't be found.  */
      errno = ESRCH;
      return -1;
    }

  /* Unlink the virtual host from the configuration file.  */
  link = host_start->prev;
  link->next = host_end->next;

  /* Remove next pointer and free the allocated memory.  */
  host_end->next = NULL;
  vsd_free_file (host_start);

  return 0;
}

/* Add a virtual host declaration at the next line after `head'.
   Perform some rudimentary syntax checking on the entry - people don't
   seem to be able to type correctly.

   Return 0 on success, -1 on error.  */
static int add_virtual_host (struct connection *vc,
			     struct vsd_file_line *head,
			     const char *data) 
{
  struct vsd_file_line *first = head, *line;
  char *p, *ibuf;
  int c, c1;

  /* Read incoming data, a line at a time.  */
  while ((ibuf = io_getline ('\n', &data)))
    {
      p = ibuf;
      /* Skip whitespace and ignore blank lines.  */
      while (*p && isspace (*p))
	p++;

      if (*p)
	{
	  /* Clean up by removing trailing spaces.  */
	  char *q = p + strlen (p) - 1;
	  while (q > p && isspace (*q))
	    q--;
	  q[1] = '\0';

	  if (vsd_file_insert_line (head, p))
	    {
	      io_buffer_store (vc, "-ERR Out of memory");
	      return -1;
	    }
	  head = head->next;
	}
      free (ibuf);
    }

  /* Perform some simple checks on the data just received.  Most critically
     we need to ensure <VirtualHost> and </VirtualHost> are the first and
     last lines and that they are spelt correctly because the scanning
     code elsewhere in this source will otherwise not find them.  */
  first = first->next;

  /* Report an error if there are multiple VirtualHost decls.  */
  c = c1 = 0;
  for (line = first; line != head; line = line->next)
    {
      if (strstr (line->line, "<VirtualHost") != NULL)
	c++;
      if (strstr (line->line, "/VirtualHost") != NULL)
	c1++;
    }
  if (c > 1 || c1 > 1)
    {
      io_buffer_store (vc, "-ERR Cannot store multiple VirtualHost declarations");
      return -1;
    }

  /* The first line should solely contain `<VirtualHost xxx>'.  */
  if (strncmp (first->line, "<VirtualHost ", 13) != 0)
    {
      io_buffer_store (vc, "-ERR The first line must be `<VirtualHost ...>'");
      return -1;
    }
  if (first->line[strlen (first->line) - 1] != '>')
    {
      io_buffer_store (vc, "-ERR <VirtualHost line has missing `>' or contains junk");
      return -1;
    }

  if (strcmp (head->line, "</VirtualHost>") != 0)
    {
      /* No </VirtualHost>. A common problem is that the browser has
	 appended it to the end of the previous line.  */
      if ((p = strstr (head->line, "</VirtualHost>")) != NULL)
	{
	  *p = '\0';
	  vsd_file_insert_line (head, "</VirtualHost>");
	}
      else
	{
	  io_buffer_store (vc, "-ERR The last line must be `</VirtualHost>'");
	  return -1;
	}
    }

  return check_vh_config (vc, first, head);
}

/* Extra features. Make sure pathname for DocumentRoot, ErrorLog
   TransferLog and CustomLog exists.

   Return 0 on success, -1 on error.  */
static int check_vh_config (struct connection *vc, struct vsd_file_line *line,
			    struct vsd_file_line *last)
{
  const char *tokens[] = { "DocumentRoot", "ErrorLog", "TransferLog",
			   "CustomLog", NULL };
  uid_t min_uid, max_uid;

  vsd_get_uidgid (vc->server_root, &min_uid, &max_uid, NULL, NULL);
  for (; line != last; line = line->next)
    {
      struct stat sb;
      char *argp, *p, *q;
      int ret, arglen, len, x;

      for (x = 0; tokens[x] && strncmp (line->line, tokens[x],
					strlen (tokens[x])) != 0; x++)
	;

      if (! tokens[x])
	continue;

      len = strlen (tokens[x]);

      /* Check there is at least one space between token and argument.  */
      if (line->line[len] != ' ' && line->line[len] != '\t')
	{
	  io_buffer_store (vc, "-ERR Need a space between directive and argument");
	  return -1;
	}

      /* Now check the file path exists.  */
      p = line->line + len;
      /* Skip spaces between directive and path.  */
      while (*p && isspace (*p))
	p++;
      /* Find end of path.  */
      q = p;
      while (*q && ! isspace (*q))
	q++;
      arglen = q - p;
      argp = (char *) malloc (arglen + 1 + strlen (vc->server_root) + 1);
      strcpy (argp, vc->server_root);
      strncat (argp, p, arglen);
      argp[arglen + strlen (vc->server_root)] = '\0';

      /* The fast case. The sensible user has already set the
	 paths up.  */
      ret = stat (argp, &sb);

      /* Chop path to virtual server root.  */
      strncpy (argp, p, arglen);
      argp[arglen] = '\0';

      if (ret == 0)
	{
	  /* Something exists but we need to ensure it is of the right
	     type. */
	  if (strcmp (tokens[x], "DocumentRoot") == 0)
	    {
	      if (! S_ISDIR (sb.st_mode))
		{
		  io_buffer_store (vc, "-ERR Document root exists but path points to a file rather than a directory");
		  return -1;
		}
	    }
	  else
	    {
	      if (S_ISDIR (sb.st_mode))
		{
		  io_buffer_store (vc, "-ERR %s path exists but points to a directory rather than a file", argp);
		  return -1;
		}
	    }
	}
      else if (*argp == '/')
	{
	  /* Currently we can only cope with absolute pathnames.  */
	  char *path;
	  uid_t uid = 0;
	  gid_t gid = 0;
	  int pathlen;

	  /* An element, or many elements, of the path do not exist.
	     Create the missing pieces and set make them owned by the
	     user of the last piece of existing directory.  We have
	     to be careful of hacker types who try and abuse the
	     system and create directories anywhere on the VS.  In this
	     case, we make sure that directories can only be created
	     within directories already owned by a user on the VS.  */
	  simplify_pathname (argp);
	  pathlen = strlen (argp);
	  path = (char *) malloc (pathlen + 1);
	  path[0] = '.'; path[1] = '\0';
	  chdir (vc->server_root);

	  /* We must ensure that directives that don't point to directories
	     should not have the final element created as a directory.  */
	  if (strcmp (tokens[x], "DocumentRoot") != 0)
	    pathlen = strrchr (argp, '/') - argp;

	  /* Walk the path.  */
	  p = argp;
	  while (*p && *p != '/')
	    p++;
	  p++;
	  while (pathlen > (p - argp))
	    {
	      /* Skip leading slashes.  */
	      q = p;
	      while (*q && *q != '/')
		q++;

	      strcat (path, "/");
	      strncat (path, p, q - p);
	      p = q + 1;
	      if (stat (path, &sb) && errno == ENOENT)
		{
		  if (uid < min_uid || uid > max_uid)
		    {
		      io_buffer_store (vc, "-ERR Cannot use path `%s': permission denied on segment `%s'", argp, path);
		    }
		  else
		    {
		      /* Create the directory and set the ownership to
			 the same as the parent directory.  */
		      mkdir (path, 0755);
		      chown (path, uid, gid);
		    }
		}
	      else
		{
		  /* Remember the uid and gid of this directory. We will
		     refer to it when creating new directories.  */
		  uid = sb.st_uid;
		  gid = sb.st_gid;
		}
	    }
	}
      free (argp);
    }
  return 0;
}

static struct vsd_file_line *find_virtual_host (struct vsd_file_line *head,
						const char *virtual_host,
						const char *servername)
{
  char *vhost, *sname;

  for (; head; head = head->next)
    if ((vhost = scan_directive (head->line,"<VirtualHost", '>')))
      if (strcmp (vhost, virtual_host) == 0)
	{
	  /* Found a matching VirtualHost. Now look for a matching
	     ServerName.  */
	  struct vsd_file_line *x = head->next;
	  int found = 0;

	  free (vhost);
	  for (x = head->next; x && ! found; x = x->next)
	    {
	      if ((sname = scan_directive (x->line, "ServerName", '\n')))
		{
		  char *p;

		  /* Strip trailing junk from the server name.  */
		  for (p = sname; *p && ! isspace (*p) && *p != '#'; p++)
		    ;
		  *p = '\0';

		  if (strcmp (servername, sname) == 0)
		    {
		      free (sname);
		      return head;
		    }

		  free (sname);
		}

	      if (! strncmp (vsd_file_skip_spaces (x->line),
			     "</VirtualHost>", 14))
		found = 1;
	    }

	  /* x is either NULL, or host_start points to a VirtualHost
	     with no ServerName.  */
	  if (found && ! strcmp (servername, "unknown"))
	    return head;
	}

  /* Couldn't find VirtualHost.  */
  errno = ENOENT;
  return NULL;
}

/* Return a list of VirtualHost declarations within httpd.conf.  */
void httpd_enumerate_virtual_hosts (struct connection *vc, int argc,
				    char *argv[])
{
  struct vsd_file_line *httpd_conf;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      int num_hosts;
      char **host_list = scan_virtual_hosts (httpd_conf, &num_hosts);

      if (num_hosts > 0)
	{
	  int x;
	  for (x = 0; x < num_hosts; x++)
	    {
	      io_buffer_store (vc, "%s\n", host_list[x]);
	      free (host_list[x]);
	    }

	  free (host_list);
	}
      vsd_free_file (httpd_conf);
    }
}

void httpd_delete_virtual_host (struct connection *vc, int argc, char *argv[])
{
  struct vsd_file_line *httpd_conf;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      if (! delete_virtual_host (httpd_conf, argv[0], argv[1]))
	vsd_save_file (httpd_conf, vc->server_root, HTTPD_CONF);
      vsd_free_file (httpd_conf);
    }
}

void httpd_add_virtual_host (struct connection *vc, int argc, char *argv[],
			     const char *data)
{
  struct vsd_file_line *httpd_conf;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      if (! add_virtual_host (vc, vsd_file_tail (httpd_conf), data))
	vsd_save_file (httpd_conf, vc->server_root, HTTPD_CONF);
      vsd_free_file (httpd_conf);
    }
}

void httpd_modify_virtual_host (struct connection *vc, int argc, char *argv[],
				const char *data)
{
  struct vsd_file_line *httpd_conf, *p;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
  else
    {
      p = find_virtual_host (httpd_conf, argv[0], argv[1]);
      if (p == NULL)
	io_buffer_store (vc, "-ERR VirtualHost '%s' not found", argv[0]);
      else
	{
	  p = p->prev;
	  if (! delete_virtual_host (httpd_conf, argv[0], argv[1]) 
	      && ! add_virtual_host (vc, p, data))
	    vsd_save_file (httpd_conf, vc->server_root, HTTPD_CONF);
	}
      vsd_free_file (httpd_conf);
    }
}

void httpd_get_virtual_host (struct connection *vc, int argc, char *argv[])
{
  struct vsd_file_line *httpd_conf, *start, *end;

  if (vsd_load_file (&httpd_conf, vc->server_root, HTTPD_CONF))
    {
      io_buffer_store (vc, "-ERR Cannot load httpd.conf: %m");
      return;
    }

  start = find_virtual_host (httpd_conf, argv[0], argv[1]);
  if (start == NULL)
    io_buffer_store (vc, "-ERR VirtualHost `%s' (%s) not found",
		     argv[0], argv[1]);
  else
    {
      /* Find the end of the VirtualHost declaration.  */
      for (end = start; end; end = end->next)
	if (strncmp (vsd_file_skip_spaces (end->line), "</VirtualHost>",
		     14) == 0)
	  break;

      if (end == NULL)
	io_buffer_store (vc, "-ERR Could not find the end of the VirtualHost declaration");
      else /* Output VirtualHost definition.  */
	for (; start != end->next; start = start->next)
	  io_buffer_store (vc, "%s\n", start->line);
    }
  vsd_free_file (httpd_conf);
}

