/* Module command interface (client interface)
   Copyright (c) 1999, 2000 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "module.h"
#include "mod_apache.h"
#include "vsd_lib.h"

struct command_tag mod_apache_command_list[] = 
{
  {
    "apache_vhenum", CMD_APACHE_VHENUM, ADMIN_VS,
    "List configured virtual hosts",
    2, "apache_vhenum <host server> <virtual server>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "apache_vhdel", CMD_APACHE_VHDELETE, ADMIN_VS,
    "Delete a virtual host definition",
    4, "apache_vhdel <host server> <virtual server> <virtual host> <server name>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "apache_vhget", CMD_APACHE_VHGET, ADMIN_VS,
    "Get configuration for a virtual host",
    4, "apache_vhget <host server> <virtual server> <virtual host> <server name>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    "apache_vhadd", CMD_APACHE_VHADD, ADMIN_VS,
    "Add a new virtual host definition",
    3, "apache_vhadd <host server> <virtual server> <filename>",
    NULL, 2, NULL /* for vsclients */
  },
  {
    "apache_vhmod", CMD_APACHE_VHMODIFY, ADMIN_VS,
    "Modify an existing virtual host",
    5, "apache_vhmod <host server> <virtual server> <virtual host> <server name> <filename>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    NULL, 0, 0,
    NULL, 0, NULL,
    NULL, 0, NULL
  }
};

/* If the command is not implemented by this module or at the level `level'
   then return `notexist'. If there are an invalid number of arguments
   then return the module syntax. If the syntax is correct then
   return NULL.  */
const char *mod_apache_check_syntax (int level, const char *command,
				     int argc, char *argv[])
{
  struct command_tag *list;

  list = module_lookup_command (mod_apache_command_list, level, command);

  /* Command doesn't exist.  */
  if (! list)
    return "notexist";

  if (level == ADMIN_CLIENT)
    {
      if (argc < list->client_command_args)
	return list->client_command_syntax;

      return NULL;
    }

  if (argc != list->num_args)
    return list->syntax;

  return NULL;
}

int mod_apache_list_help (int level, const char *command)
{
  struct command_tag *list;

  if (command)
    {
      list = module_lookup_command (mod_apache_command_list, level, command);
      if (list == NULL)
	return 1;

      fprintf (stderr, "  %14s - %s\n", list->tag, list->help);
    }
  else
    {
      int y = 0;

      list = mod_apache_command_list;
      if (list == NULL)
	return 1;

      while (list[y].tag)
	{
	  if (list[y].level & level)
	    fprintf (stderr, "  %14s - %s\n", list[y].tag, list[y].help);
	  y++;
	}
    }

  return 0;
}

char *mod_apache_do_cmd (vsd_server *server, int level,
			 const char *cmd, char **argv)
{
  struct command_tag *list;
  char *status;

  list = module_lookup_command (mod_apache_command_list, level, cmd);
  if (! list)
    return "notexist";

  server->level = list->level;
  switch (list->token)
    {
    case CMD_APACHE_VHENUM:
      vsd_send_command (server, NULL, "APACHE_VHENUM");
      status = vsd_process_reply (server);
      break;
    case CMD_APACHE_VHDELETE:
      vsd_send_command (server, NULL, "APACHE_VHDELETE \"%s\" \"%s\"",
			argv[1] /* virtual host */,
			argv[2] /* server name */);
      status = vsd_process_reply (server);
      break;
    case CMD_APACHE_VHGET:
      vsd_send_command (server, NULL, "APACHE_VHGET \"%s\" \"%s\"",
			argv[1] /* virtual host */,
			argv[2] /* server name */);
      status = vsd_process_reply (server);
      break;
    case CMD_APACHE_VHADD:
      vsd_send_command_file (server, argv[1], "APACHE_VHADD %s");
      status = vsd_process_reply (server);
      break;
    case CMD_APACHE_VHMODIFY:
      vsd_send_command_file (server, argv[3], "APACHE_VHMODIFY \"%s\" \"%s\"",
			     argv[1] /* virtual host */,
			     argv[2] /* server name */);
      status = vsd_process_reply (server);
      break;
    default:
      status = "mod_apache: command not recognised";
      break;
    }

  return status;
}

module apache_module =
{
  mod_apache_check_syntax,
  mod_apache_list_help,
  mod_apache_do_cmd
};
