/* Module command interface (client interface)
   Copyright (c) 1999, 2000 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "module.h"
#include "mod_priv.h"
#include "vsd_lib.h"

struct command_tag mod_priv_command_list[] = 
{
  {
    "priv_modify", CMD_PRIV_MODIFY, ADMIN_VS,
    "Set privileges for a user on a virtual server",
    3, "priv_modify <host server> <virtual server> <user> <+|-><privilege>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    "priv_list", CMD_PRIV_LIST, ADMIN_VS,
    "List privileges assigned to a user",
    3, "priv_list <host server> <virtual server> <user>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    "priv_avail", CMD_PRIV_AVAIL, ADMIN_VS,
    "List available privileges on a virtual server",
    2, "priv_avail <host server> <virtual server>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    NULL, 0, 0,
    NULL, 0, NULL,
    NULL, 0, NULL
  }
};

/* If the command is not implemented by this module or at the level `level'
   then return `notexist'. If there are an invalid number of arguments
   then return the module syntax. If the syntax is correct then
   return NULL.  */
const char *mod_priv_check_syntax (int level, const char *command,
				   int argc, char *argv[])
{
  struct command_tag *list;

  list = module_lookup_command (mod_priv_command_list, level, command);

  /* Command doesn't exist.  */
  if (! list)
    return "notexist";

  if (level == ADMIN_CLIENT)
    {
      if (argc < list->client_command_args)
	return list->client_command_syntax;

      return NULL;
    }

  if (argc != list->num_args)
    return list->syntax;

  return NULL;
}

int mod_priv_list_help (int level, const char *command)
{
  struct command_tag *list;

  if (command)
    {
      list = module_lookup_command (mod_priv_command_list, level, command);
      if (list == NULL)
	return 1;

      fprintf (stderr, "  %14s - %s\n", list->tag, list->help);
    }
  else
    {
      int y = 0;

      list = mod_priv_command_list;
      if (list == NULL)
	return 1;

      while (list[y].tag)
	{
	  if (list[y].level & level)
	    fprintf (stderr, "  %14s - %s\n", list[y].tag, list[y].help);
	  y++;
	}
    }

  return 0;
}

char *mod_priv_do_cmd (vsd_server *server, int level,
		       const char *cmd, char **argv)
{
  struct command_tag *list;
  char *status;

  list = module_lookup_command (mod_priv_command_list, level, cmd);
  if (! list)
    return "notexist";

  server->level = list->level;
  switch (list->token)
    {
    case CMD_PRIV_MODIFY:
      vsd_send_command (server, NULL, "PRIV_MODIFY %s %s", argv[1] /* user */,
			argv[2] /* priv */);
      status = vsd_process_reply (server);
      break;
    case CMD_PRIV_LIST:
      if (argv[1])
	vsd_send_line (server, "PRIV_LIST %s", argv[1]);
      else
	vsd_send_line (server, "PRIV_LIST");

      status = vsd_process_reply (server);
      break;
    case CMD_PRIV_AVAIL:
      vsd_send_line (server, "PRIV_AVAIL");
      status = vsd_process_reply (server);
      break;
    default:
      status = "mod_priv: command not recognised";
      break;
    }

  return status;
}

module priv_module =
{
  mod_priv_check_syntax,
  mod_priv_list_help,
  mod_priv_do_cmd
};
