/* Module command interface (client interface)
   Copyright (c) 1999, 2000 Idaya Ltd.
   Written by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "module.h"
#include "mod_sendmail.h"
#include "vsd_lib.h"

struct command_tag mod_sendmail_command_list[] = 
{
  {
    "sendmail_vlist", CMD_SENDMAIL_VLIST, ADMIN_VS,
    "List e-mail aliases for a domain",
    3, "sendmail_vlistdomain <host server> <virtual server> <domain>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "sendmail_vmodify", CMD_SENDMAIL_VMODIFY, ADMIN_VS,
    "Modify e-mail aliases for a domain",
    3, "sendmail_vmoddomain <host server> <virtual server> <filename>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "sendmail_cwlist", CMD_SENDMAIL_CWLIST, ADMIN_VS,
    "List domains for which we receive mail",
    2, "sendmail_cwlist <host server> <virtual server>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    "sendmail_cwadd", CMD_SENDMAIL_CWADD, ADMIN_VS,
    "Add a domain for which we want to receive mail",
    4, "sendmail_cwadd <host server> <virtual server> <user> <domain name>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    "sendmail_cwdelete", CMD_SENDMAIL_CWDELETE, ADMIN_VS,
    "Delete a domain name from sendmail.cw",
    3, "sendmail_cwdelete <host server> <virtual server> <domain name>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    NULL, 0, 0,
    NULL, 0, NULL,
    NULL, 0, NULL
  }
};

/* If the command is not implemented by this module or at the level `level'
   then return `notexist'. If there are an invalid number of arguments
   then return the module syntax. If the syntax is correct then
   return NULL.  */
const char *mod_sendmail_check_syntax (int level, const char *command,
				       int argc, char *argv[])
{
  struct command_tag *list;

  list = module_lookup_command (mod_sendmail_command_list, level, command);

  /* Command doesn't exist.  */
  if (! list)
    return "notexist";

  if (level == ADMIN_CLIENT)
    {
      if (argc < list->client_command_args)
	return list->client_command_syntax;

      return NULL;
    }

  if (argc != list->num_args)
    return list->syntax;

  return NULL;
}

int mod_sendmail_list_help (int level, const char *command)
{
  struct command_tag *list;

  if (command)
    {
      list = module_lookup_command (mod_sendmail_command_list, level, command);
      if (list == NULL)
	return 1;

      fprintf (stderr, "  %14s - %s\n", list->tag, list->help);
    }
  else
    {
      int y = 0;

      list = mod_sendmail_command_list;
      if (list == NULL)
	return 1;

      while (list[y].tag)
	{
	  if (list[y].level & level)
	    fprintf (stderr, "  %14s - %s\n", list[y].tag, list[y].help);
	  y++;
	}
    }

  return 0;
}

char *mod_sendmail_do_cmd (vsd_server *server, int level,
			   const char *cmd, char **argv)
{
  struct command_tag *list;
  char *status;

  list = module_lookup_command (mod_sendmail_command_list, level, cmd);
  if (! list)
    return "notexist";

  server->level = list->level;
  switch (list->token)
    {
    case CMD_SENDMAIL_VLIST:
      vsd_send_command (server, NULL, "SENDMAIL_VLIST %s",
			argv[3] /* domain */); 
      status = vsd_process_reply (server);
      break;
    case CMD_SENDMAIL_VMODIFY:
      vsd_send_command (server, NULL, "SENDMAIL_VMODIFY");
      vsd_send_file (server, argv[3] /* filename */);
      status = vsd_process_reply (server);
      break;
    case CMD_SENDMAIL_CWLIST:
      vsd_send_command (server, NULL, "SENDMAIL_CWLIST"); 
      status = vsd_process_reply (server);
      break;
    case CMD_SENDMAIL_CWADD:
      vsd_send_command (server, NULL, "SENDMAIL_CWADD %s",
			argv[3] /* domain */); 
      status = vsd_process_reply (server);
      break;
    case CMD_SENDMAIL_CWDELETE:
      vsd_send_command (server, NULL, "SENDMAIL_CWDELETE %s",
			argv[3] /* domain */); 
      status = vsd_process_reply (server);
      break;
    default:
      status = "mod_sendmail: command not recognised";
      break;
    }

  return status;
}

module sendmail_module =
{
  mod_sendmail_check_syntax,
  mod_sendmail_list_help,
  mod_sendmail_do_cmd
};
