/* Sendmail configuration file manipulation.
   Copyright (c) 1999, 2000 Idaya Ltd.
   Written by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

#include "vsd.h"
#include "libvsd.h"
#include "sendmail.h"

/* Read /etc/sendmail.cf to get the paths for virtusertable, aliases
   and sendmail.cw.   If sendmail.cf contains the relevant information
   then allocate some memory and store the paths in the appropriate
   variable.  Otherwise, store NULL in the pointer.  */
static int read_sendmail_cf (const char *server_root,
			     char **virtusertable, char **aliases,
			     char **sendmail_cw)
{
  struct vsd_file_line *head, *line;
  char buf[256];

  if (virtusertable)
    *virtusertable = NULL;
  if (aliases)
    *aliases = NULL;
  if (sendmail_cw)
    *sendmail_cw = NULL;

  if (vsd_load_file (&head, server_root, "/etc/sendmail.cf"))
    return;

  line = head;
  while (line)
    {
      if (aliases && strncasecmp (line->line, "O AliasFile", 11) == 0)
	{
	  /* O AliasFile=/etc/aliases  */
	  char *p = strchr (line->line, '=');
	  if (p != NULL)
	    {
	      p++;
	      while (isspace (*p))
		p++;
	      *aliases = (char *) malloc (strlen (p) + 1);
	      strcpy (*aliases, p);
	    }
	}
      if (virtusertable && strncasecmp (line->line, "Kvirtuser", 9) == 0)
	{
	  /* Kvirtuser hash -o /etc/virtusertable.  */
	  char *p = strstr (line->line, "-o");
	  char *q;
	  if (p != NULL)
	    {
	      /* Skip past the -o and any spaces.  */
	      p += 2;
	      while (isspace (*p))
		p++;

	      q = p;
	      while (! isspace (*q))
		q++;

	      *virtusertable = (char *) malloc ((q - p) + 2);
	      strncpy (*virtusertable, p, q - p + 1);
	      (*virtusertable)[q - p + 1] = '\0';
	    }
	}
      if (sendmail_cw && strncasecmp (line->line, "Fw", 2) == 0)
	{
	  /* Fw/etc/sendmail.cw  */
	  *sendmail_cw = (char *) malloc (strlen (line->line + 2) + 1);
	  strcpy (*sendmail_cw, line->line + 2);
	}
      line = line->next;
    }

  return 0;
}

/* /etc/virtusertable manipulation.

   The syntax of /etc/virtusertable is:
     <e-mail alias><space|tab><pop3 account>

   Lines beginning with a # are comments.  */

struct virt_entry
{
  char *alias;
  char *pop_account;
  struct virt_entry *next;
};

/* Add an entry to the virtusertable linked list.
   Return 0 on success, -1 on failure.  */
static int virt_add_entry (struct virt_entry **list,
			   const char *alias, const char *pop)
{
  struct virt_entry *entry;

  entry = (struct virt_entry *) malloc (sizeof (struct virt_entry));
  if (entry == NULL)
    return -1;

  entry->alias = (char *) malloc (strlen (alias) + 1);
  entry->pop_account = (char *) malloc (strlen (pop) + 1);
  if (entry->alias == NULL || entry->pop_account == NULL)
    return -1;

  strcpy (entry->alias, alias);
  strcpy (entry->pop_account, pop);
  entry->next = NULL;

  if (*list == NULL)
    *list = entry;
  else
    {
      struct virt_entry *llist = *list;

      /* Add entry to end of list.  */
      while (llist->next)
	llist = llist->next;

      llist->next = entry;
    }

  return 0;
}

/* Delete an entry from the virtusertable linked list.
   Return 0 on success, -1 on failure.  */
static int virt_delete_entry (struct virt_entry **list,
			      const char *alias, const char *pop)
{
  struct virt_entry *prev, *entry;

  prev = NULL;
  entry = *list;
  while (entry)
    {
      /* Find a matching entry.  */
      if (strcmp (entry->alias, alias) == 0
	  && strcmp (entry->pop_account, pop) == 0)
	{
	  if (prev == NULL)
	    {
	      /* Entry is first in the list.  */
	      prev = entry->next;
	      free (entry->alias);
	      free (entry->pop_account);
	      free (entry);
	      *list = prev;
	      return 0;
	    }

	  /* Unlink entry from middle of the list.  */
	  prev->next = entry->next;
	  free (entry->alias);
	  free (entry->pop_account);
	  free (entry);
	  return 0;
	}
      prev = entry;
      entry = entry->next;
    }

  /* No matching entry.  */
  errno = ENOENT;
  return -1;
}
	  
/* Load /etc/virtusertable into memory and store in a linked list.
   Return a pointer to the head of the linked list on success,
   or NULL on error and set errno.

   The values of errno are:
      ESRCH - Could not find the virtual server root directory
              in /etc/passwd.
      ENOENT - /etc/virtusertable does not exist
      ENOSPC - there are no entries in /etc/virtusertable.  */
static struct virt_entry *load_virtusertable (struct connection *vc,
					      const char *server)
{
  FILE *stream;
  char line[512], *p, alias[256], pop_account[256], *fname;
  struct virt_entry *list;
  int x;

  read_sendmail_cf (vc->server_root, &fname, NULL, NULL);
  if (fname == NULL)
    return NULL;
  
  sprintf (line, "%s%s", vc->server_root, fname);
  stream = fopen (line, "r");
  if (stream == NULL)
    return NULL;

  free (fname);
  list = NULL;
  while (fgets (line, sizeof (line), stream))
    {
      /* Skip lines with comments.  */
      p = line;
      while (isspace (*p))
	p++;
      if (*p == '#' || *p == '\0')
	continue;

      /* Obtain the e-mail alias.  */
      x = 0;
      while (!isspace (*p) && x < sizeof (alias))
	alias[x++] = *p++;
      alias[x] = '\0';

      /* Skip white space. */
      while (isspace (*p))
	p++;

      /* Obtain the pop3 account.  */
      x = 0;
      while (*p && *p != '\n' && *p != '\r' && x < sizeof (pop_account))
	pop_account[x++] = *p++;
      pop_account[x] = '\0';

      if (virt_add_entry (&list, alias, pop_account))
	break;
    }

  fclose (stream);

  if (list == NULL)
    errno = ENOSPC;

  return list;
}

/* Save the virtusertable, stored in the linked `list' to /etc/virtusertable
   on the virtual server `server'.
   Return 0 on success, -1 on error.  */
static int save_virtusertable (struct connection *vc, struct virt_entry *list,
			       const char *server)
{
  FILE *stream;
  char line[256], *fname;

  read_sendmail_cf (vc->server_root, &fname, NULL, NULL);
  if (fname == NULL)
    return -1;

  sprintf (line, "%s%s", vc->server_root, fname);
  free (fname);
  stream = fopen (line, "w");
  if (stream == NULL)
    return -1;

  while (list)
    {
      fprintf (stream, "%s %s\n", list->alias, list->pop_account);
      list = list->next;
    }

  fclose (stream);

  return 0;
}

static int rebuild_virtusertable_db (struct connection *vc, const char *server)
{
  char *fname, *cmd;
  int status;

  read_sendmail_cf (vc->server_root, &fname, NULL, NULL);
  if (fname)
    {
      cmd = (char *) malloc (strlen (fname) + 9);
      sprintf (cmd, "hash %s.db", fname);
    }

  status = run_command (vc, vc->server_root, "admin",
			"/usr/sbin/makemap", cmd, fname, NULL);
  free (cmd);
  free (fname);
  return status;
}

/* Free the memory allocated to the virtusertable structure.  */
static void free_virtusertable (struct virt_entry *list)
{
  struct virt_entry *temp;

  while (list)
    {
      temp = list->next;
      free (list);
      list = temp;
    }
}

/* Comparison function for qsort.  */
static int virt_compare (const void *a, const void *b)
{
  const struct virt_entry *e1 = (const struct virt_entry *) a;
  const struct virt_entry *e2 = (const struct virt_entry *) b;
  char *d1, *d2;
  int cmp;

  d1 = strchr (e1->alias, '@');
  d2 = strchr (e2->alias, '@');

  /* Might be trying to process an invalid entry.  */
  if (d1 == NULL)
    d1 = e1->alias;
  if (d2 == NULL)
    d2 = e2->alias;

  cmp = strcmp (d1 + 1, d2 + 1);
  if (cmp == 0)
    {
      /* Comparing the same domain.  */

      /* Check for the catch all line. Return as a greater than to ensure
	 it always ends up at the bottom of the domain list.  */
      if (e1->alias[0] == '@' && e2->alias[0] == '@')
	return 0;
      if (e1->alias[0] == '@')
	return 1;
      if (e2->alias[0] == '@')
	return 1;

      /* Otherwise return that they are the same.  */
      return strcmp (e1->alias, e2->alias);
    }

  return cmp;
}

/* Sort the virtusertable by the domain name in the alias column
   and ensure the catch-all clause comes last.  */
static void sort_virtusertable (struct virt_entry **list)
{
  struct virt_entry *array, *entry;
  int elements, x;

  /* Count the number of elements.  */
  for (elements = 0, entry = *list; entry; elements ++, entry = entry->next)
    ;
  if (! elements)
    return;
  elements --;

  /* Build the linked list into an array to be sorted by qsort.  */
  array = (struct virt_entry *) malloc ((elements + 1)
					* sizeof (struct virt_entry));
  for (x = 0, entry = *list; x <= elements; x++, entry = entry->next)
    array[x] = *entry;

  /* Sort the list.  */
  qsort (array, elements, sizeof (struct virt_entry), virt_compare);

  /* Free the old list and create a new (sorted) list.  */
  free_virtusertable (*list);
  *list = NULL;
  for (x = 0; x <= elements; x++)
    virt_add_entry (list, array[x].alias, array[x].pop_account);

  free (array);
}

/* List the e-mail addresses associated with `domain'.  */
void sendmail_virt_list_domain (struct connection *vc, int argc, char *argv[])
{
  struct virt_entry *list, *entry;
  char *p;

  list = load_virtusertable (vc, vc->virtual_server);
  if (list == NULL)
    {
      if (errno == ENOSPC)
	{
	  /* Nothing in table. Not a problem.  */
	}
      else if (errno == ESRCH)
	io_buffer_store (vc, "-ERR Virtual server %s does not exist",
			 vc->virtual_server);
      else
	io_buffer_store (vc, "-ERR Fatal error accessing virtusertable: %m");
      return;
    }

  entry = list;
  while (entry)
    {
      p = strchr (entry->alias, '@');
      /* Check for invalid entries.  */
      if (p != NULL && strstr (p, argv[0]) != NULL)
	io_buffer_store (vc, "alias=\"%s\" address=\"%s\"\n",
			 entry->alias, entry->pop_account);
      entry = entry->next;
    }

  free_virtusertable (list);
}

void sendmail_virt_modify_table (struct connection *vc, int argc, char *argv[],
				 const char *data)
{
  struct virt_entry *list;
  char *ibuf;

  list = load_virtusertable (vc, vc->virtual_server);
  if (list == NULL && errno != ENOSPC)
    {
      if (errno == ESRCH)
	io_buffer_store (vc, "-ERR Virtual server %s does not exist",
			 vc->virtual_server);
      else
	io_buffer_store (vc, "-ERR Fatal error accessing virtusertable: %m");
      return;
    }

  /* Read entries from the client.  */
  while ((ibuf = io_getline ('\n', &data)))
    {
      int vecc;
      char **vecv = vsd_argv_parse (' ', ibuf, &vecc);

      /* add <alias> <address>  */
      /* delete <alias> <address>  */
      if (strcmp (vecv[0], "add") == 0)
	virt_add_entry (&list, vecv[1], vecv[2]);
      else if (strcmp (vecv[0], "delete") == 0)
	virt_delete_entry (&list, vecv[1], vecv[2]);
      free (ibuf);
      free (vecv);
    }

  sort_virtusertable (&list);
  save_virtusertable (vc, list, vc->virtual_server);
  free_virtusertable (list);
  rebuild_virtusertable_db (vc, vc->virtual_server);
}

void sendmail_cw_list_domain (struct connection *vc, int argc, char *argv[])
{
  struct vsd_file_line *head, *entry;
  char *fname;

  read_sendmail_cf (vc->server_root, NULL, NULL, &fname);
  if (fname == NULL)
    {
      io_buffer_store (vc, "-ERR Cannot find `Fw' in sendmail.cf");
      return;
    }

  if (vsd_load_file (&head, vc->server_root, fname))
    io_buffer_store (vc, "-ERR Could not load sendmail.cw: %m");
  else
    {
      for (entry = head; entry; entry = entry->next)
	{
	  const char *p = vsd_file_skip_spaces (entry->line);
      
	  /* Skip lines with comments.  */
	  if (*p == '#')
	    continue;
	  
	  /* Skip empty lines.  */
	  if (*p == '\0')
	    continue;
  
	  /* Anything else is assumed to be a domain name that the virtual
	     server receives mail for.  */
	  io_buffer_store (vc, "%s\n", p);
	}

      vsd_free_file (head);
    }
  free (fname);
}

/* Add a domain name to /etc/sendmail.cw.  */
void sendmail_cw_add_domain (struct connection *vc, int argc, char *argv[])
{
  struct vsd_file_line *head;
  char *fname;

  read_sendmail_cf (vc->server_root, NULL, NULL, &fname);
  if (fname == NULL)
    {
      io_buffer_store (vc, "-ERR Cannot find `Fw' in sendmail.cf");
      return;
    }

  if (vsd_load_file (&head, vc->server_root, fname))
    io_buffer_store (vc, "-ERR Could not load sendmail.cw: %m");
  else
    {
      /* Insert domain name into /etc/sendmail.cw.  */
      if (vsd_file_insert_line (vsd_file_tail (head), argv[0]))
	io_buffer_store (vc, "-ERR Could not add domain: %m");
      else if (vsd_save_file (head, vc->server_root, fname))
	io_buffer_store (vc, "-ERR Could not save sendmail.cw: %m");

      vsd_free_file (head);
    }
  free (fname);
}

/* Delete a domain name from /etc/sendmail.cw.  */
void sendmail_cw_delete_domain (struct connection *vc, int argc,
				char *argv[])
{
  struct vsd_file_line *head, *entry;
  char *fname;

  read_sendmail_cf (vc->server_root, NULL, NULL, &fname);
  if (fname == NULL)
    {
      io_buffer_store (vc, "-ERR Cannot find `Fw' in sendmail.cf");
      return;
    }

  if (vsd_load_file (&head, vc->server_root, fname))
    io_buffer_store (vc, "-ERR An error occurred: %m");
  else
    {
      /* Find domain in /etc/sendmail.cw.  */
      entry = vsd_file_find_line (head, argv[0]);
      if (entry == NULL)
	io_buffer_store (vc, "-ERR Domain name not found");
      else
	{
	  /* Delete domain from /etc/sendmail.cw.  */
	  head = vsd_file_delete_line (head, entry);

	  if (vsd_save_file (head, vc->server_root, fname))
	    io_buffer_store (vc, "-ERR Could not save sendmail.cw: %m");
	}
      vsd_free_file (head);
    }

  free (fname);
}


void sendmail_alistlhs (struct connection *vc, int argc, char *argv[])
{
#if 0
  struct vsd_file_line *head, *entry;
  char *fname;

  read_sendmail_cf (vc->server_root, NULL, &fname, NULL);
  if (fname == NULL)
    {
      io_buffer_store (vc, "-ERR Cannot find `O AliasFile' in sendmail.cf");
      return NULL;
    }

  if (vsd_load_file (&head, vc->server_root, fname))
    {
      io_buffer_store (vc, "-ERR Cannot load aliases file: %m");
      return;
    }

  vsd_free_file (head);
  free (fname);
#endif
}

void sendmail_agetrhs (struct connection *vc, int argc, char *argv[])
{
#if 0
  struct vsd_file_line *head, *entry;
  char *fname;

  read_sendmail_cf (vc->server_root, NULL, &fname, NULL);
  if (fname == NULL)
    {
      io_buffer_store (vc, "-ERR Cannot find `O AliasFile' in sendmail.cf");
      return NULL;
    }

  if (vsd_load_file (&head, vc->server_root, fname))
    {
      io_buffer_store (vc, "-ERR Cannot load aliases file: %m");
      return;
    }

  vsd_free_file (head);
#endif
}

void sendmail_aadd (struct connection *vc, int argc, char *argv[])
{
#if 0
  struct vsd_file_line *head;
  char *fname;

  read_sendmail_cf (vc->Server_root, NULL, &fname, NULL);
  if (fname == NULL)
    {
      io_buffer_store (vc, "-ERR Cannot find `O AliasFile' in sendmail.cf");
      return NULL;
    }

  if (vsd_load_file (&head, vc->server_root, fname))
    {
      io_buffer_store (vc, "-ERR Cannot load aliases file: %m");
      return;
    }
#endif
}

void sendmail_adelete (struct connection *vc, int argc, char *argv[])
{
#if 0
  struct vsd_file_line *head, *entry;
  char *fname;

  read_sendmail_cf (vc->server_root, NULL, &fname, NULL);
  if (fname == NULL)
    {
      io_buffer_store (vc, "-ERR Cannot find `O AliasFile' in sendmail.cf");
      return NULL;
    }

  if (vsd_load_file (&head, vc->server_root, fname))
    {
      io_buffer_store (vc, "-ERR Cannot load aliases file: %m");
      return;
    }

#endif
}

