/* Module command interface (client interface)
   Copyright (c) 1999, 2000 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "module.h"
#include "mod_user.h"
#include "vsd_lib.h"

struct command_tag mod_user_command_list[] = 
{
  {
    "user_add", CMD_USER_ADD, ADMIN_VS,
    "Add a user",
    3, "user_add <host server> <virtual server> <user name>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "user_del", CMD_USER_DEL, ADMIN_VS,
    "Delete a user",
    3, "user_del <host server> <virtual server> <user name>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "user_mod", CMD_USER_MODIFY, ADMIN_VS,
    "Modify an existing user",
    3, "user_mod <host server> <virtual server> <user name>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "user_list", CMD_USER_LIST, ADMIN_VS,
    "List all users on a virtual server",
    2, "user_list <host server> <virtual server>",
    NULL, 0, NULL  /* for vsclients  */
  },
  {
    "user_chpass", CMD_USER_CHPASS, ADMIN_VS,
    "Change password for a user",
    5, "user_chpass <host server> <virtual server> <user> <old pass> <new pass>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "user_gradd", CMD_USER_GRADD, ADMIN_VS,
    "Create a new group",
    3, "user_add <host server> <virtual server> <group name>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "user_grdel", CMD_USER_GRDEL, ADMIN_VS,
    "Delete an existing group",
    3, "user_add <host server> <virtual server> <group name>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "user_grmod", CMD_USER_GRMOD, ADMIN_VS,
    "Add/remove a user from a group",
    4, "user_grmod <host server> <virtual server> <group name> <+|-><user>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "user_grlist", CMD_USER_LIST, ADMIN_VS,
    "List all user groups on a virtual server",
    2, "user_grlist <host server> <virtual server>",
    NULL, 0, NULL  /* for vsclients  */
  },
  {
    NULL, 0, 0,
    NULL, 0, NULL,
    NULL, 0, NULL
  }
};

/* If the command is not implemented by this module or at the level `level'
   then return `notexist'. If there are an invalid number of arguments
   then return the module syntax. If the syntax is correct then
   return NULL.  */
const char *mod_user_check_syntax (int level, const char *command,
				   int argc, char *argv[])
{
  struct command_tag *list;

  list = module_lookup_command (mod_user_command_list, level, command);

  /* Command doesn't exist.  */
  if (! list)
    return "notexist";

  if (level == ADMIN_CLIENT)
    {
      if (argc < list->client_command_args)
	return list->client_command_syntax;

      return NULL;
    }

  if (argc != list->num_args)
    return list->syntax;

  return NULL;
}

int mod_user_list_help (int level, const char *command)
{
  struct command_tag *list;

  if (command)
    {
      list = module_lookup_command (mod_user_command_list, level, command);
      if (list == NULL)
	return 1;

      fprintf (stderr, "  %14s - %s\n", list->tag, list->help);
    }
  else
    {
      int y = 0;

      list = mod_user_command_list;
      if (list == NULL)
	return 1;

      while (list[y].tag)
	{
	  if (list[y].level & level)
	    fprintf (stderr, "  %14s - %s\n", list[y].tag, list[y].help);
	  y++;
	}
    }

  return 0;
}

char *mod_user_do_cmd (vsd_server *server, int level,
		       const char *cmd, char **argv)
{
  struct command_tag *list;
  char *status;

  list = module_lookup_command (mod_user_command_list, level, cmd);
  if (! list)
    return "notexist";

  server->level = list->level;
  switch (list->token)
    {
    case CMD_USER_ADD:
      vsd_send_command (server, NULL, "USER_ADD %s", argv[2] /* user */);
      status = vsd_process_reply (server);
      break;
    case CMD_USER_DEL:
      vsd_send_command (server, NULL, "USER_DEL %s", argv[2] /* user */);
      status = vsd_process_reply (server);
      break;
    case CMD_USER_MODIFY:
      vsd_send_command (server, NULL, "USER_MOD %s", argv[2] /* user */);
      status = vsd_process_reply (server);
      break;
    case CMD_USER_LIST:
      vsd_send_command (server, NULL, "USER_LIST");
      status = vsd_process_reply (server);
      break; 
   case CMD_USER_CHPASS:
      vsd_send_command (server, NULL, "USER_CHPASS %s %s %s",
			argv[2] /* user */, argv[3] /* old pass */,
			argv[4] /* new pass */);
      status = vsd_process_reply (server);
      break;
    case CMD_USER_GRADD:
      vsd_send_command (server, NULL, "USER_GRADD %s", argv[2] /* group */);
      status = vsd_process_reply (server);
      break;
    case CMD_USER_GRDEL:
      vsd_send_command (server, NULL, "USER_GRDEL %s", argv[2] /* group */);
      status = vsd_process_reply (server);
      break;
    case CMD_USER_GRMOD:
      vsd_send_command (server, NULL, "USER_GRMOD %s", argv[2] /* group */,
			argv[3] /* +|- user */);
      status = vsd_process_reply (server);
      break;
    case CMD_USER_GRLIST:
      vsd_send_command (server, NULL, "USER_GRLIST");
      status = vsd_process_reply (server);
      break;
    default:
      status = "mod_user: command not recognised";
      break;
    }

  return status;
}

module user_module =
{
  mod_user_check_syntax,
  mod_user_list_help,
  mod_user_do_cmd
};
