/* Module command interface (client interface)
   Copyright (c) 1999, 2000 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include "module.h"
#include "mod_vs.h"
#include "vsd_lib.h"

struct command_tag mod_vs_command_list[] = 
{
  {
    "vs_create", CMD_VS_CREATE, ADMIN_HOST,
    "Create a new virtual server",
    6, "vs_create <host server> <virtual server> <ip-addr> <fqdn> <numuid> <quota>",
    NULL, 0, NULL
  },
  {
    "vs_delete", CMD_VS_DELETE, ADMIN_HOST,
    "Delete a virtual server",
    2, "vs_delete <host server> <virtual server>",
    NULL, 0, NULL
  },
  {
    "vs_enable", CMD_VS_ENABLE, ADMIN_HOST,
    "Enable a virtual server",
    2, "vs_enable <host server> <virtual server>", NULL, 0, NULL
  },
  {
    "vs_disable", CMD_VS_DISABLE, ADMIN_HOST,
    "Disable a virtual server",
    2, "vs_disable <host server> <virtual server>",
    NULL, 0, NULL
  },
  {
    "vs_details", CMD_VS_DETAILS, ADMIN_HOST,
    "Get virtual server setup details",
    1, "vs_details <host server> [<virtual server>]",
    NULL, 0, NULL
  },
  {
    "vs_reboot", CMD_VS_REBOOT, ADMIN_VS | ADMIN_CLIENT,
    "Restart services on a virtual server",
    2, "vs_reboot <host server> <virtual server>",
    "rebootvs", 0, ""
  },
  {
    "vs_signal", CMD_VS_SIGNAL, ADMIN_VS,
    "Send a signal to all processes on a virtual server",
    3, "vs_signal <host server> <virtual server> <signal number>",
    NULL, 0, NULL
  },
  {
    "vs_modify", CMD_VS_SIGNAL, ADMIN_HOST,
    "Change characteristics of a virtual server",
    7, "vs_modify <host server> <old-vs> <new-vs> <old-fqdn> <new-fqdn> <old-ip> <new-ip>",
    NULL, 0, NULL
  },
  {
    NULL, 0, 0,
    NULL, 0, NULL,
    NULL, 0, NULL
  }
};

/* If the command is not implemented by this module or at the level `level'
   then return `notexist'. If there are an invalid number of arguments
   then return the module syntax. If the syntax is correct then
   return NULL.  */
const char *mod_vs_check_syntax (int level, const char *command,
				 int argc, char *argv[])
{
  struct command_tag *list;

  list = module_lookup_command (mod_vs_command_list, level, command);

  /* Command doesn't exist.  */
  if (! list)
    return "notexist";

  if (level == ADMIN_CLIENT)
    {
      if (argc != list->client_command_args)
	return list->client_command_syntax;

      return NULL;
    }

  if (argc < list->num_args)
    return list->syntax;

  return NULL;
}

int mod_vs_list_help (int level, const char *command)
{
  struct command_tag *list;

  if (command)
    {
      list = module_lookup_command (mod_vs_command_list, level, command);
      if (list == NULL)
	return 1;

      fprintf (stderr, "  %14s - %s\n", list->tag, list->help);
    }
  else
    {
      int y = 0;

      list = mod_vs_command_list;
      if (list == NULL)
	return 1;

      while (list[y].tag)
	{
	  if (list[y].level & level)
	    fprintf (stderr, "  %14s - %s\n", list[y].tag, list[y].help);
	  y++;
	}
    }

  return 0;
}

char *mod_vs_do_cmd (vsd_server *server, int level,
		     const char *cmd, char **argv)
{
  struct command_tag *list;
  char *status;

  list = module_lookup_command (mod_vs_command_list, level, cmd);
  if (! list)
    return "notexist";

  server->level = list->level;
  switch (list->token)
    {
    case CMD_VS_CREATE:
      vsd_send_command (server, NULL, "VS_CREATE %s %s %s %s %s",
			argv[1] /* vs */, argv[2] /* ipaddr */,
			argv[3] /* fqdn */, argv[4] /* numuid */,
			argv[5] /* quota */);
      status = vsd_process_reply (server);
      break;
    case CMD_VS_DELETE:
      vsd_send_command (server, NULL, "VS_DELETE %s", argv[1] /* vs */);
      status = vsd_process_reply (server);
      break;
    case CMD_VS_ENABLE:
      vsd_send_command (server, NULL, "VS_ENABLE %s", argv[1] /* vs */);
      status = vsd_process_reply (server);
      break;
    case CMD_VS_DISABLE:
      vsd_send_command (server, NULL, "VS_DISABLE %s", argv[1] /* vs */);
      status = vsd_process_reply (server);
      break;
    case CMD_VS_DETAILS:
      if (argv[1])
	vsd_send_command (server, NULL, "VS_DETAILS %s", argv[1] /* vs */);
      else
	vsd_send_command (server, NULL, "VS_DETAILS");
      status = vsd_process_reply (server);
      break;
    case CMD_VS_REBOOT:
      vsd_send_command (server, NULL, "VS_REBOOT");
      status = vsd_process_reply (server);
      break;
    case CMD_VS_SIGNAL:
      vsd_send_command (server, NULL, "VS_SIGNAL %d", argv[1] /* vs */);
      status = vsd_process_reply (server);
      break;
    case CMD_VS_MODIFY:
      vsd_send_command (server, NULL, "VS_MODIFY %s %s %s %s %s %s",
			argv[1] /* old-vs */, argv[2] /* new-vs */,
			argv[3] /* old-fqdn */, argv[4] /* new-fqdn */,
			argv[5] /* old-ip */, argv[6] /* new-ip */);
      status = vsd_process_reply (server);
      break;
    default:
      status = "mod_vs: command not recognised";
      break;
    }

  return status;
}

module vs_module =
{
  mod_vs_check_syntax,
  mod_vs_list_help,
  mod_vs_do_cmd
};

