/* Command line interface for the virtual server administrator.
   Copyright (c) 1999, 2000 Idaya Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdarg.h>
#include <sys/syslog.h>
#include <unistd.h>
#include <netdb.h>
#include <pwd.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <sys/utsname.h>
#ifdef HAVE_SYS_SYSCALL_H
#include <sys/syscall.h>
#endif
#ifdef HAVE_ASM_UNISTD_H
/* _syscall might be defined in here.  */
#include <asm/unistd.h>
#endif

#include "libvsd.h"
#include "module.h"

char *program_name;

extern char *stpcpy (char *dest, const char *src);
static void error (int usage,
		   const char *fmt, ...) __attribute__ ((__noreturn__));

/* #define DEBUG 1 */

/* If no data has been read in READ_TIMEOUT seconds then timeout.  */
#define READ_TIMEOUT 20

/* Establish a tcp connection to the virtual server administrator
   on host 'host'.
   Return the socked descriptor on success, or -1 on failure.  */
static int establish_connection (const char *host)
{
  struct hostent *hostent;
  struct sockaddr_in addr;
  int sock;

  hostent = gethostbyname (host);
  if (hostent == NULL)
    {
      printf ("could not resolve host %s\n", host);
      return -1;
    }

  memset (&addr, 0, sizeof (addr));
  addr.sin_family = AF_INET;

  addr.sin_port = htons (WITH_VSD_PORT);
  addr.sin_addr = *(struct in_addr *) hostent->h_addr;

  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock == -1)
    {
      printf ("socket create failed: %s\n", strerror (errno));
      return -1;
    }

  if (connect (sock, (struct sockaddr *) &addr, sizeof (addr)) == -1)
    {
      printf ("connection to %s could not be established: %s\n",
	      host, strerror (errno));
      return -1;
    }

  return sock;
}

/* Read a line from the virtual server.
   Return a pointer to the line on success or NULL on failure.  */
char *vsd_get_line (vsd_server *vsd)
{
  char *cp;
  time_t start_time, end_time;

  /* Does any data exist in the buffer ? */
  if (vsd->data)
    {
      /* Look for a newline character in the buffer.  */
      cp = strchr (vsd->buffer + vsd->dp, '\n');

      if (cp)
	{
	  int found;

	  /* A newline was found.  */
	  *cp = '\0'; /* Terminate the string.  */
	  found = vsd->dp; /* Mark the start of the string.  */
	  /* Reduce the amount of data left in the buffer.  */
	  vsd->data -= (cp + 1 - (vsd->buffer + vsd->dp));

	  /* Move the data pointer to the start of the next string.  */
	  vsd->dp = (cp + 1) - vsd->buffer;

#ifdef DEBUG
	  printf ("get_line: '%s'\n", vsd->buffer + found);
#endif
	  return vsd->buffer + found;
	}
      else
	{
	  /* No newline was found. Move remaining data to the beginning
	     of the buffer.  */
	  memmove (vsd->buffer, vsd->buffer + vsd->dp, vsd->data);
	  vsd->dp = 0; /* Reset line index to the beginning.  */
	}
    }
  else
    vsd->dp = 0;

  if (! vsd->buffer)
    {
      /* Initially allocate 256 bytes for the receive buffer.  */
      vsd->buffer_size = 256;
      vsd->buffer = (char *) malloc (vsd->buffer_size + 1);
    }

  start_time = time (0);
  while (1)
    /*  while (vsd->data < vsd->buffer_size) */
    {
      int ret;

#ifdef DEBUG
      printf ("-- buffer_size = %d, data = %d\n", 
	      vsd->buffer_size, vsd->data);
#endif

      if (vsd->buffer_size - vsd->data < 16)
	{
#ifdef DEBUG
	  printf ("increase buffer from %d by 256\n", vsd->buffer_size);
#endif
	  /* Getting tight on memory, increase the read buffer size.  */
	  vsd->buffer_size += 256;
	  vsd->buffer = (char *) realloc (vsd->buffer, vsd->buffer_size + 1);
	}

      /* Retry read if failure was from an interrupt. */
      memset (vsd->buffer + vsd->data, 0, vsd->buffer_size - vsd->data);
#ifdef WITH_OPENSSL
      ret = SSL_read (vsd->ssl, &vsd->buffer[vsd->data],
		      vsd->buffer_size - vsd->data);
#else
      while ((ret = read (vsd->fd, &vsd->buffer[vsd->data],
			  vsd->buffer_size - (vsd->data + 1))) == -1
	     && errno == EINTR)
	;
#endif

#ifdef DEBUG
      printf ("-- ret = %d, buffer = ", ret);
      for (x = 0; x <= ret; x++)
        printf ("%c", vsd->buffer[vsd->data + x]);
      printf ("[the end]\n");
#endif

      if (ret == 0)
	{
	  struct timespec period, remain;
	  /* Timeout if read failure for READ_TIMEOUT seconds.  */
	  end_time = time (0);
	  if (end_time - start_time > READ_TIMEOUT)
	    {
	      printf ("Read timeout\n");
	      return NULL;
	    }
	  /* Sleep for 0.2 seconds to stop read timeouts taking up
	     too much CPU.  */
	  period.tv_sec = 0;
	  period.tv_nsec = 200000000;
	  nanosleep (&period, &remain);
	}

      vsd->data += ret;
      cp = strchr (vsd->buffer, '\n');
      if (cp)
	{
	  *cp = '\0';
	  vsd->data -= (cp + 1 - (vsd->buffer + vsd->dp));
	  vsd->dp = (cp + 1) - vsd->buffer;

#ifdef DEBUG
	  printf ("get_line: '%s'\n", vsd->buffer);
#endif
	  return vsd->buffer;
	}
    }

  return NULL;
}

/* Send 'line' to the virtual server administrator.
   Return 0 on success, -1 on failure.  */
int vsd_send_line (vsd_server *server, const char *fmt, ...)
{
  va_list ap;
  int ret, len, nchars, size;
  char *line;

  va_start (ap, fmt);

  size = 256;
  line = (char *) malloc (size);
  nchars = vsnprintf (line, size, fmt, ap);
  if (nchars >= size)
    {
      size = nchars + 2;
      line = (char *) realloc (line, size);
      vsnprintf (line, size, fmt, ap);
    }

#ifdef DEBUG
  printf ("vsd_send_line: '%s'\n", line);
#endif

  strcat (line, "\n");
  len = strlen (line);

  /* Retry write if failure is due to an interrupt.  */
  while ((ret = write (server->fd, line, len)) == -1
	 && errno == EINTR)
    ;

  free (line);
  if (ret == len)
    return 0;

  return -1;
}

/* Open a connection to the virtual server adminstrator hosted on
   'host_server'. Connect as user 'username' and password 'password'.
   If 'virtual_server' is NULL then perform operations on the hosting
   server only, otherwise perform operations on the virtual server.

   'username' will be an existing standard user on the host server,
   if virtual_server is NULL, or a standard user virtaul server if
   `virtual_server' is non NULL.

   Return a malloced pointer to a vsd_server structure if a connection
   is successfully established.  Return NULL if an error occurred.  */
vsd_server *vsd_open (const char *host_server, const char *virtual_server,
		      const char *username, const char *password)
{
  int sock;
  vsd_server *server;

#ifdef DEBUG
  printf ("vsd_open: host_server='%s' virtual_server='%s' username='%s'\n",
	  host_server, virtual_server, username);
#endif

  if (host_server == NULL)
    {
      errno = EINVAL;
      return NULL;
    }

  /* Determine user name.  */
  if (username == NULL)
    {
      username = getenv ("USER");
      if (username == NULL || *username == '\0')
	{
	  username = getlogin ();
	  if (username == NULL || *username == '\0')
	    {
	      struct passwd *pwd;

	      pwd = getpwuid (getuid ());
	      if (pwd != NULL && pwd->pw_name != NULL
		  && pwd->pw_name[0] != '\0')
		username = pwd->pw_name;
	      else
		return NULL;
	    }
	}
    }

  server = (vsd_server *) malloc (sizeof (vsd_server));
  if (server == NULL)
    {
      printf ("Memory allocation error\n");
      return NULL;
    }

  errno = 1;
  /* Connect to the virtual server administrator.  */
  sock = establish_connection (host_server);
  if (sock == -1)
    {
      free (server);
      return NULL;
    }

  server->fd = sock;
  server->data = 0;
  server->buffer = NULL;
  server->dp = 0;

  /* The virtual server administrator will immediately send us a +OK
     message when a connection has been established.  Check for it here
     and close the socket if it doesn't appear.  */
  if (vsd_reply_success (server))
    {
      close (sock);
      free (server);
      return NULL;
    }

  if (virtual_server == NULL)
    strcpy (server->virtual_server, "<root>");
  else
    strcpy (server->virtual_server, virtual_server);

  /* Connection successfully established and we are now logged on.  */
  return server;
}

/* Close the connection to the virtual server administrator.  Send a QUIT
   message, giving the administrator the opportunity to tidy up.  */
void vsd_close (vsd_server *server)
{
#ifdef DEBUG
  printf ("vsd_close: closing connection\n");
#endif
  close (server->fd);
  free (server);
}

/* Receive a reply from the virtual server administrator.
   If it is a +OK message then return 0.
   If it is a -ERR message, return 1 (i.e. failure).  */
int vsd_reply_success (vsd_server *server)
{
  char *line;

  line = vsd_get_line (server);
  if (line == NULL)
    return 1;

  if (strncmp (line, "+OK", 3) == 0)
    return 0;

  /* Error messages are of the form: -ERR<space><error msg>  */
  printf ("%s\n", line + sizeof ("-ERR"));
  return 1;
}

/* Standard function for those that receive data but don't do much else
   with it.  */
char *vsd_process_reply (vsd_server *server)
{
  char *line, *in;
  char **vec = NULL;
  size_t vecc = 0;
  dstring_t buf = dstring_new (32);
  int level = 0;
  int flag = 0;

  while (1)
    {
      line = vsd_get_line (server);
      if (line == NULL)
	break;

      in = line;
      if (level != 2)
	{
	  /* Skip blanks.  */
	  while (*in && isspace (*in))
	    in++;
	  if (! *in)
	    continue;
	  if (in[0] == 'E' && in[1] == 'O' && in[2] == 'F')
	    break;

	  /* Split arguments into an argument vector.  */
	  vec = vsd_argv_parse (' ', in, &vecc);
	  if (vec == NULL)
	    continue;
	}

      if (level == 0)
	{
	  /* Expect `virtual server {'.  */
	  level ++;
	  free (vec);
	  flag = 1;
	}
      else if (level == 1)
	{
	  /* Expect a command or a `}'.  */
	  if (*in == '}')
	    {
	      level --;
	    }
	  else
	    {
	      if (*vec[vecc - 1] == '{')
		{
		  level ++;
		  vecc --;
		}
	    }
	}
      else if (level == 2)
	{
	  if (strchr (in, '}'))
	    {
	      /* End of data segment.  */
	      level --;
	    }
	  else
	    dstring_fmt_append (buf, "%s\n", in);
	}
    }

  if (buf->length)
    {
      char *s = strdup (buf->s);
      dstring_free (buf);
      return s;
    }

  if (! flag)
    return "-ERR expected result but got nothing";

  return NULL;
}

int vsd_send_file (vsd_server *server, const char *file)
{
  FILE *stream;
  char line[256];

  stream = fopen (file, "r");
  if (stream != NULL)
    {
      while (fgets (line, sizeof (line), stream))
	vsd_send_line (server, "+DATA %s", line);

      fclose (stream);
    }
  return vsd_send_line (server, "+END");
}

int vsd_send_command (vsd_server *server,
		      const char *data, const char *fmt, ...)
{
  va_list ap;
  int status, nchars, size;
  char *line;

  va_start (ap, fmt);
  status = vsd_send_line (server, "%s {",
			  ((server->level & ADMIN_HOST)
			   ? "<root>" : server->virtual_server));

  size = 256;
  line = (char *) malloc (size);
  nchars = vsnprintf (line, size, fmt, ap);
  if (nchars >= size)
    {
      size = nchars + 2;
      line = (char *) realloc (line, size);
      vsnprintf (line, size, fmt, ap);
    }

  if (data)
    status = vsd_send_line (server, "%s {\n%s\n}", line, data);
  else
    status = vsd_send_line (server, "%s", line);
  status = vsd_send_line (server, "}\nEOF");

  free (line);
  va_end (ap);
  return status;
}

int vsd_send_command_file (vsd_server *server,
			   const char *fname, const char *fmt, ...)
{
  va_list ap;
  int status, nchars, size;
  char *line;

  va_start (ap, fmt);
  status = vsd_send_line (server, "%s {", server->virtual_server);

  size = 256;
  line = (char *) malloc (size);
  nchars = vsnprintf (line, size, fmt, ap);
  if (nchars >= size)
    {
      size = nchars + 2;
      line = (char *) realloc (line, size);
      vsnprintf (line, size, fmt, ap);
    }

  if (fname)
    {
      FILE *stream = fopen (fname, "r");

      status = vsd_send_line (server, "%s {\n", line);
      if (stream != NULL)
	{
	  while (fgets (line, sizeof (line), stream))
	    vsd_send_line (server, "%s", line);

	  fclose (stream);
	}
      vsd_send_line (server, "}\n");
    }
  else
    status = vsd_send_line (server, "%s", line);
  status = vsd_send_line (server, "}\nEOF");
  free (line);

  va_end (ap);
  return status;
}

/* Return the host server name.
   Return -1 on failure, 0 on success.  */
int vsd_gethostservername (char *buffer, int size)
{
  struct utsname buf;
  size_t node_len;
#ifdef linux
  extern int ___uname (struct utsname *);

  _syscall1 (int, uname, struct utsname *, buf);

  if (buffer == NULL)
    {
      errno = EINVAL;
      return -1;
    }
#endif

  /* `gethostname' and `uname' have been hacked to return the FQDN of
     the virtual server.  Fortunately, as a result of the hack, we are
     left with `___uname' which will give us the information we require.  */

  if (uname (&buf))
    return -1;

  node_len = strlen (buf.nodename) + 1;
  if (node_len > size)
    {
      errno = ENAMETOOLONG;
      return -1;
    }

  strcpy (buffer, buf.nodename);
  return 0;
}

static void error (int usage, const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  if (usage == 1)
    fprintf (stdout, "usage: %s ", program_name);

  vfprintf (stdout, fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);

  exit (1);
}

int main (int argc, char *argv[])
{
  vsd_server *server;
  char host_server[256], virtual_server[256];
  char *status;
  char *operation = argv[1];
  const char *st;

  openlog ("vsdadm", LOG_PID, LOG_DAEMON);
  if (operation == NULL || argc <= 1)
    {
      module_help (ADMIN_VS | ADMIN_HOST, NULL);
      return 1;
    }

  switch (argc)
    {
    case 2:
      syslog (LOG_WARNING, "cmd: %s", argv[1]);
      break;
    case 3:
      syslog (LOG_WARNING, "cmd: %s %s", argv[1], argv[2]);
      break;
    case 4:
      syslog (LOG_WARNING, "cmd: %s %s %s", argv[1], argv[2], argv[3]);
      break;
    case 5:
      syslog (LOG_WARNING, "cmd: %s %s %s %s", argv[1], argv[2], argv[3],
	      argv[4]);
      break;
    case 6:
      syslog (LOG_WARNING, "cmd: %s %s %s %s %s", argv[1],
	      argv[2], argv[3], argv[4], argv[5]);
      break;
    case 7:
      syslog (LOG_WARNING, "cmd: %s %s %s %s %s %s", argv[1],
	      argv[2], argv[3], argv[4], argv[5], argv[6]);
      break;
    case 8:
      syslog (LOG_WARNING, "cmd: %s %s %s %s %s %s %s", argv[1],
	      argv[2], argv[3], argv[4], argv[5], argv[6], argv[7]);
      break;
    }
  
  program_name = argv[0];

  /* Might be a module.  */
  st = module_syntax (ADMIN_VS | ADMIN_HOST, operation, argc - 2, argv + 1);
  if (st != NULL)
    {
      if (strcmp (st, "notexist") == 0)
	error (0, "command %s does not exist", operation);
      else
	error (1, "%s", st);
    }

  /* Get the host server and virtual server names.  */
  strcpy (host_server, argv[2]);
  if (argc >= 4)
    strcpy (virtual_server, argv[3]);
  else
    strcpy (virtual_server, "<root>");
  server = vsd_open (host_server, virtual_server, "root", "NOAUTH");
  if (server == NULL)
    return 1;

  /* Perform the action on the virtual server.  */
  status = module_command (server, ADMIN_VS | ADMIN_HOST, operation, argv + 2);
  if (status && strncmp (status, "-ERR", 4) == 0)
    {
      fprintf (stdout, "error: %s\n", status);
      return 1;
    }

  if (status)
    fprintf (stdout, "%s\n", status);
  return 0;
}

